# Copyright 2021 Karlsruhe Institute of Technology
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from flask_babel import gettext as _
from flask_babel import lazy_gettext as _l

from kadi.lib.config.core import get_sys_config
from kadi.lib.config.core import set_sys_config
from kadi.lib.conversion import empty_str
from kadi.lib.forms import BooleanField
from kadi.lib.forms import FileField
from kadi.lib.forms import JSONField
from kadi.lib.forms import KadiForm
from kadi.lib.forms import LFTextAreaField
from kadi.lib.forms import SubmitField


class NavFooterField(JSONField):
    """Custom field to process and validate navigation footer items.

    Only performs some basic validation to make sure the overall structure of the items
    is valid.
    """

    def _value(self):
        return self.data if self.data else []

    def process_formdata(self, valuelist):
        super().process_formdata(valuelist)

        if valuelist:
            if not isinstance(self.data, list):
                raise ValueError(_("Invalid data."))

            for item in self.data:
                if not isinstance(item, list) or len(item) != 2:
                    raise ValueError(_("Invalid data."))
        else:
            self.data = []


class BaseConfigForm(KadiForm):
    """Base form class for use in setting global config items.

    All fields in an inheriting form will be populated automatically from suitable
    config items stored in the database, if applicable. As keys for the config items,
    the field names are taken in uppercase.

    :param ignored_fields: (optional) A set of field names, as specified in the class
        attributes, to ignore when prepopulating the form.
    """

    submit = SubmitField(_l("Save changes"))

    def __init__(self, *args, ignored_fields=None, **kwargs):
        self.ignored_fields = ignored_fields if ignored_fields is not None else set()
        self.ignored_fields.add("submit")

        data = {}

        def _prepopulate_fields(field_name):
            config_value = get_sys_config(field_name.upper(), use_fallback=False)

            if config_value is not None:
                data[field_name] = config_value

        self._iterate_fields(_prepopulate_fields)

        super().__init__(*args, data=data, **kwargs)

    def _iterate_fields(self, callback):
        """Iterate through all fields that are not ignored.

        :param callback: A callback which will be invoked for each field with its name.
        """
        for field in self._unbound_fields:
            field_name = field[0]

            if field_name not in self.ignored_fields:
                callback(field_name)

    def set_config_values(self):
        """Automatically set all global config items based on the respective field data.

        Useful to populate all relevant config items in the database after a form is
        submitted. Similar to prepopulating the form fields, the names of the fields are
        taken in uppercase as keys for each config item.

        Note that fields that are ignored when prepopulating will be ignored in this
        case as well.
        """
        self._iterate_fields(
            lambda field_name: set_sys_config(
                field_name.upper(), getattr(self, field_name).data
            )
        )


class CustomizationConfigForm(BaseConfigForm):
    """A form for use in setting global config items related to customization."""

    broadcast_message = LFTextAreaField(
        _l("Broadcast message"),
        filters=[empty_str],
        description=_l("Shown at the top of all pages to all logged-in users."),
    )

    nav_footer_items = NavFooterField(
        _l("Navigation footer items"),
        description=_l(
            "Shown on all pages in the footer next to the existing navigation items."
        ),
    )

    index_image = FileField(
        _l("Index image"),
        description=_l("Shown on the index page next to the index text."),
    )

    remove_image = BooleanField(_l("Remove current image"))

    index_text = LFTextAreaField(
        _l("Index text"),
        filters=[empty_str],
        description=_l("Shown on the index page next to the index image."),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, ignored_fields={"index_image", "remove_image"}, **kwargs
        )


class MiscConfigForm(BaseConfigForm):
    """A form for use in setting miscellaneous global config items."""

    robots_noindex = BooleanField(
        _l("Do not index website"),
        description=_l(
            "Exclude this website from being indexed by all major search engines."
        ),
    )
